/* MiscTreeBrowser
 *
 */

// IMPORTS

#import <appkit/Text.h>

#import <misckit/MiscTreeBrowser.h>
#import <misckit/MiscTreeBrowserCell.h>
#import <misckit/MiscTreeBrowserMatrix.h>
#import <misckit/MiscTreeBrowserProtocols.h>


// DEFINES

#define DEFAULT_BG_RED	0.80079
#define DEFAULT_BG_GRN	0.73438
#define DEFAULT_BG_BLU	0.66797

#define DEFAULT_BG_COLOR	\
	NXConvertRGBToColor(DEFAULT_BG_RED, DEFAULT_BG_GRN, DEFAULT_BG_BLU)


// CLASS VARIABLES

static NXColor defaultBackgroundColor;


// IMPLEMENTATION

@implementation MiscTreeBrowser

// PRIVATE CLASS METHODS

+ initialize
{
	if ([MiscTreeBrowser class] == self)
	{
		defaultBackgroundColor = DEFAULT_BG_COLOR;
	}

	return self;
}

// PRIVATE METHODS

- (BOOL)_node:(id<MiscTreeBrowserNodes>)aNode
  isDescendantOf:(id<MiscTreeBrowserNodes>)parentNode
{
	id node = aNode;

	while ((node = [node parent]))
	{
		if (node == parentNode)
		{
			return YES;
		}
	}
	return NO;
}

/*"
Inserts a cell into the matrix at index, containing node.  Does not call -sizeToCells or -display to redisplay the matrix, allowing rapid insertion of more than one cell e.g. opening a node with children.
"*/
- _insertCellWithNode:node at:(unsigned int)index
{
	unsigned int cellCount = [cellMatrix cellCount];
	id cell;

	if (index > cellCount)
	{							// if cell beyond end of matrix, change
		index = cellCount;		// to just at end of the matrix
	}
								// add cell to matrix
	[cellMatrix insertRowAt:index];
	cell = [cellMatrix cellAt:index :0];
								// initialise cell
	[cell setState:0];
	[cell setOpen:NO];
	[cell setNode:node];
	[cell setTarget:self];
	[cell setDoubleAction:@selector(doubleClick:)];

	return self;
}


/*"
Index of next cell following aCell whose node is not a descendant of [aCell node].  Used for e.g. adding a sibling.
"*/
- (int)_indexAfterNodeInCell:aCell
{
	int idx = -1;

	return idx;
}

- (int)_nextNonAncestorOf:cell after:(int)index
{
	int cellCount = [cellMatrix cellCount];
	id parentNode = [cell node];

	while (++index < cellCount)
	{
		id node = [[cellMatrix cellAt:index :0] node];

		if (![self _node:node isDescendantOf:parentNode])
		{
			return index;
		}
	}

	return index;
}

// PUBLIC METHODS

- mouseDragged:(NXEvent *)theEvent
{
//	printf("MiscTreeView:  mouse drag\n");

	return self;
}

- doubleClick:sender
{
//	printf("MiscTreeView:  double click\n");

	return self;
}

- init
{
  return [self initFrame:NULL];
}

- initFrame:(const NXRect *)frameRect
{
	NXSize interCellSpacing = {0.0, 0.0}, docSize;

	[super initFrame:frameRect];

	cellMatrix = [[MiscTreeBrowserMatrix alloc] initFrame:frameRect
											mode:NX_LISTMODE
											cellClass:[MiscTreeBrowserCell class]
											numRows:0
											numCols:1];

	/*
	 * In the following lines,
	 * remember that "cellMatrix" is the matrix that will be installed
	 * in the scrollview, "self" is the scrollview.
	 */

	[self setBackgroundColor:defaultBackgroundColor];

	/* we don't want any space between the matrix's cells  */
	[cellMatrix setIntercell:&interCellSpacing];
	/* resize the matrix to contain the cells */
	[cellMatrix sizeToCells];
	[cellMatrix setAutosizeCells:YES];
	/*
	 * when the user clicks in the matrix and then drags the mouse out of
	 * scrollView's contentView, we want the matrix to scroll
	 */
	[cellMatrix setAutoscroll:YES];
	/* let the matrix stretch horizontally */
	[cellMatrix setAutosizing:NX_WIDTHSIZABLE];  
	/* Install the matrix as the docview of the scrollview */
	[[self setDocView:cellMatrix] free];
	/* set up the visible attributes of the scrollview */
	[self setVertScrollerRequired:YES];
	[self setBorderType:NX_BEZEL];
	/* tell the subviews to resize along with the scrollview */
	[self setAutoresizeSubviews:YES];
	/* This is the only way to get the clipview to resize too */
	[[cellMatrix superview] setAutoresizeSubviews:YES];
	/* Allow the scrollview to stretch both horizontally and vertically */
	[self setAutosizing:NX_WIDTHSIZABLE|NX_HEIGHTSIZABLE];
	/* Resize the matrix to fill the inside of the scrollview */
	[self getContentSize:&docSize];
	[cellMatrix sizeTo:docSize.width :docSize.height];

	return self;
}

- free
{
	[cellMatrix free];
	return [super free];
}

- cellMatrix
{
	return cellMatrix;
}

- setFont:aFont
{
	[cellMatrix setFont:aFont];

	return self;
}

- font
{
	return [cellMatrix font];
}

- setRoot:anObject
{
	MiscTreeBrowserCell *cell;

	/* tell the matrix there are 0 cells in it (but don't deallocate them) */
	[cellMatrix renewRows:0 cols:1];
	[cellMatrix lockFocus];		/* for highlightCellAt::lit: */

	/*
	 * add a row to the matrix.  (This doesn't necessarily allocate a new
	 * cell, thanks to renewRows:cols:).
	 */
	[cellMatrix addRow];
	cell = [cellMatrix cellAt:0:0];
	/* make sure the cell is neither selected nor highlighted */
	[cellMatrix highlightCellAt:0:0 lit:NO];
	[cell setState:0];
	/* install the node in that cell */
	[cell setNode:anObject];
	[cell setDoubleAction:@selector(doubleClick:)];

	[cellMatrix unlockFocus];
	[cellMatrix sizeToCells];
	[cellMatrix display];
  
	return self;
}

- addRoot:aNode
{
	[self addRoot:aNode atIndex:[cellMatrix cellCount] + 1];

	return self;
}

- addRoot:aNode atIndex:(unsigned int)index
{
	[aNode setParent:nil];
    [self _insertCellWithNode:aNode at:index];
	[cellMatrix sizeToCells];
	[cellMatrix display];

	return self;
}

- addChild:aNode
{
    id cell = [cellMatrix selectedCell];

	if (cell)
	{
		[aNode setParent:[cell node]];
		if ([cell isOpen])
		{
			[self _insertCellWithNode:aNode at:[cellMatrix selectedRow] + 1];
		}
		[cellMatrix sizeToCells];
		[cellMatrix display];
	}
	else
	{
		[self addRoot:aNode];
	}

	return self;
}

- addChild:anObject atIndex:(unsigned int)index
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];

	return self;
}

- addSibling:aNode
{
    id cell = [cellMatrix selectedCell];

	if (cell)
	{
		[aNode setParent:[[cell node] parent]];
		if ([cell isOpen])
		{
			[self _insertCellWithNode:aNode at:[cellMatrix selectedRow] + 1];
		}
		[cellMatrix sizeToCells];
		[cellMatrix display];
	}
	else
	{
		[self addRoot:aNode];
	}

	return self;
}

- addSibling:anObject atIndex:(unsigned int)index
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];
	return self;
}

- removeNode
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];
	return self;
}


- removeCellAt:(int)index
{
	id cell = [cellMatrix cellAt:index :0];

	if (cell)
	{
		id<MiscTreeBrowserNodes> node = [cell node];

		if (node && [node canFree])
		{
			id<MiscTreeBrowserNodes> parent = [node parent];

			if ([cell isOpen] && ![cell isLeaf])
			{
				[self closeCell:cell];
			}
			if (parent)
			{
				[parent removeChild:node];
			}
			else
			{
				[node free];
			}
			[cell setNode:nil];		
			[cellMatrix  removeRowAt:index andFree:NO];
			[cellMatrix sizeToCells];
			[cellMatrix display];
		}
	}

	return self;
}

- addCellAt:(int)index
{
	id cell = [cellMatrix cellAt:index :0];

	if (cell)
	{
		id<MiscTreeBrowserNodes> node = [cell node];

		if (node)
		{
			id<MiscTreeBrowserNodes> newSib = [node createSibling];

			if (newSib)
			{
				int sibIndex = [self _nextNonAncestorOf:cell after:index];
				[self _insertCellWithNode:newSib at:sibIndex];
				[cellMatrix sizeToCells];
				[cellMatrix display];
				[cellMatrix editCell:[cellMatrix cellAt:sibIndex :0]];
			}
		}
	}

	return self;
}

- removeNodeAtIndex:(unsigned int)index
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];
	return self;
}

- reloadFromNode:aNode
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];
	return self;
}

- reload
{
    [self error:"  %s objects should not be sent '%s' messages\n",
            [[self class] name], sel_getName(_cmd)];
	return self;
}

// Opening and closing cells

/*"
"*/
- openCell:aCell
{
	int row, col;

	if ([aCell isOpen])
	{							// do nothing if already open
		return self;
	}
	[cellMatrix getRow:&row andCol:&col ofCell:aCell];

	if (row >= 0)
	{
		id node = [aCell node];
		int numChildren = [node childCount];

		[aCell setOpen:YES];
		if (numChildren > 0)
		{
			int i;

			for (i = 0; i < numChildren; i++)
			{
				row++;
				[self _insertCellWithNode:[node childAt:i] at:row];
			}
			[cellMatrix sizeToCells];
		}
		[cellMatrix display];
	}

	return self;
}

/*"
"*/
- closeCell:aCell
{
	int row, col;

	if (![aCell isOpen])
	{							// do nothing if already closed
		return self;
	}
	[cellMatrix getRow:&row andCol:&col ofCell:aCell];
	if (row >= 0)
	{
		int cellCount = [cellMatrix cellCount];
		id thisNode = [aCell node];

		row++;
		while ((row < cellCount) &&
			   ([self _node:[[cellMatrix cellAt:row :0] node]
					  isDescendantOf:thisNode]))
		{
			[[cellMatrix cellAt:row :0] setNode:nil];
			[cellMatrix removeRowAt:row andFree:NO];
			cellCount--;
		}
		[aCell setOpen:NO];
		[cellMatrix sizeToCells];
		[cellMatrix display];
	}

	return self;
}

@end
